@@ -65,17 +65,27 @@ DB* gettdb(const abl_info *info) {
     return info->utdb;
 }
 
-int dbopen(const abl_args *args, const char *dbfile, const char *dbname, DB **db) {
+int dbopen(const abl_args *args, const char *dbfile, const char *dbname, DB_ENV *dbenv, DB **db) {
     int err;
-
-    if (err = db_create(db, NULL, 0), 0 != err) {
+    if (err = db_create(db, dbenv, 0), 0 != err) {
         log_db_error(err, "creating database object");
         return err;
     }
-
-    if (err = (*db)->open(*db, NULL, dbfile, dbname, DB_BTREE,
-                       DB_CREATE, DBPERM), 0 != err) {
+    
+    DB_TXN *tid = 0;
+    if ((err = dbenv->txn_begin(dbenv, NULL, &tid, 0)) != 0) {
+        log_db_error(err, "starting transaction");
+        return err;
+    }
+    
+    if (err = (*db)->open(*db, tid, dbfile, dbname, DB_BTREE, DB_CREATE, DBPERM), 0 != err) {
         log_db_error(err, "opening or creating database");
+        tid->abort(tid);
+        return err;
+    }
+    
+    if ((err = tid->commit(tid, 0))) {
+        log_db_error(err, "committing transaction");
         return err;
     }
 
@@ -84,14 +94,40 @@ int dbopen(const abl_args *args, const char *dbfile, const char *dbname, DB **db
     return 0;
 } 
 
+int dbenvironment(const char *home, DB_ENV **dbenv) {
+    int ret = 0;
+    if ((ret = db_env_create(dbenv, 0)) != 0) {
+    	log_db_error(ret, "creating environment object");
+        return ret;
+    }
+    (*dbenv)->set_errpfx(*dbenv, "pam-abl");
+#if DB_VERSION_MAJOR > 4 || (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR > 3)
+    if ((ret = (*dbenv)->open(*dbenv, home, DB_CREATE | DB_INIT_TXN | DB_INIT_LOCK | DB_INIT_MPOOL | DB_RECOVER | DB_REGISTER, 0)) != 0) {
+#else
+    if ((ret = (*dbenv)->open(*dbenv, home, DB_CREATE | DB_INIT_TXN | DB_INIT_LOCK | DB_INIT_MPOOL | DB_RECOVER, 0)) != 0) {
+#endif
+
+        log_db_error(ret, "opening the database environment");
+        (*dbenv)->close(*dbenv, 0);
+        *dbenv = 0;
+    }
+    if (*dbenv) {
+        /* Do deadlock detection internally. */
+        if ((ret = (*dbenv)->set_lk_detect(*dbenv, DB_LOCK_DEFAULT)) != 0) {
+            log_db_error(ret, "setting lock detection.");
+        }
+    }
+    return ret;
+}
+
 /* db is an already opened database.  */
-int dbget(DB *db, DBT *key, DBT *dbtdata) {
+int dbget(DB *db, DBT *key, DB_TXN *tid, DBT *dbtdata) {
     int err; 
 
     memset(dbtdata, 0, sizeof(*dbtdata));
     dbtdata->flags = DB_DBT_USERMEM;
 
-    err = db->get(db, NULL, key, dbtdata, 0);
+    err = db->get(db, tid, key, dbtdata, DB_RMW);
 
     /*Called with DB_DBT_USERMEM?  What was there wasn't enough*/
     if (TOOSMALL == err) {
@@ -100,7 +136,7 @@ int dbget(DB *db, DBT *key, DBT *dbtdata) {
         }
         dbtdata->size = 0;
         /* ...and try again. */
-        err = db->get(db, NULL, key, dbtdata, 0);
+        err = db->get(db, tid, key, dbtdata, DB_RMW);
     }
 
     if (0 != err && DB_NOTFOUND != err) {
@@ -117,7 +153,7 @@ const char * get_kv(const abl_info *info) {
     return info->subject == HOST ? info->host : info->user;
 }
     
-int update_status(const abl_args *args, abl_info *info) {
+int update_status(const abl_args *args, abl_info *info, DB_TXN *tid) {
     DB *db;
     DBT key, data;
     int err = 0;
@@ -131,14 +167,14 @@ int update_status(const abl_args *args, abl_info *info) {
     if (err) return err;
 
     make_key(&key, get_kv(info));
-    err = dbget(db, &key, &data);
+    err = dbget(db, &key, tid, &data);
     if (err && err != DB_NOTFOUND) return err;
     if (data.data == NULL || *(int *) data.data != info->state ) {
         log_debug(args,"state changed data %d info %d", data.data ? *(int *) data.data : -1,info->state);
-        info->state_change = 1;
+        info->state_change = data.data != NULL;
         data.data = &info->state;
         data.size = sizeof(info->state);
-        err = db->put(db,NULL,&key,&data,0);
+        err = db->put(db,tid,&key,&data,0);
     }
     else {
         info->state_change = 0;
@@ -163,8 +199,14 @@ int record(const abl_args *args, abl_info *info, time_t tm, long maxage) {
 
     make_key(&key, kv);
 
-    err = dbget(db, &key, &data);
+    DB_TXN *tid;
+    if ((err = info->dbenv->txn_begin(info->dbenv, NULL, &tid, 0)) != 0) {
+        log_db_error(err, "starting transaction");
+        return err;
+    }
+    err = dbget(db, &key, tid, &data);
     if (0 != err && DB_NOTFOUND != err) {
+        tid->abort(tid);
         return err;
     }
 
@@ -175,6 +217,7 @@ int record(const abl_args *args, abl_info *info, time_t tm, long maxage) {
     }
 
     if (err = grow_buffer(&data, data.size + sizeof(time_t)), 0 != err) {
+        tid->abort(tid);
         return err;
     }
 
@@ -182,16 +225,22 @@ int record(const abl_args *args, abl_info *info, time_t tm, long maxage) {
     data.size += sizeof(time_t);
 
 
-    if (err = db->put(db, NULL, &key, &data, 0), 0 != err) {
+    if (err = db->put(db, tid, &key, &data, 0), 0 != err) {
         log_sys_error(err, "updating database");
     }
+    if (err) {
+        tid->abort(tid);
+    } else {
+        if ((err = tid->commit(tid, 0)))
+            log_db_error(err, "comitting transaction.");
+    }
 
     return err;
 }
 
 /* Check whether access should be permitted for the specified user/host.
  */
-int check(const abl_args *args, abl_info *info, time_t tm) {
+int check(const abl_args *args, abl_info *info, DB_TXN *tid, time_t tm) {
     DB *db;
     DBT key, data;
     int err, sz;
@@ -213,7 +262,7 @@ int check(const abl_args *args, abl_info *info, time_t tm) {
 
     memset(&data, 0, sizeof(data));
 
-    err = dbget(db, &key, &data);
+    err = dbget(db, &key, tid, &data);
     if (0 != err) {
         if (DB_NOTFOUND == err) {
             info->state = CLEAR;
@@ -241,8 +290,24 @@ int check_host(const abl_args *args, abl_info *info, time_t tm) {
 
         if (NULL != info->host) {
             log_debug(args, "Checking host %s", info->host);
-            check(args, info, tm);
-            update_status(args, info);
+            DB_TXN *tid;
+            int err = 0;
+            if ((err = info->dbenv->txn_begin(info->dbenv, NULL, &tid, 0)) != 0) {
+                log_db_error(err, "starting transaction");
+                return err;
+            }
+            err = check(args, info, tid, tm);
+            if (err) {
+                log_debug(args, "host check");
+                tid->abort(tid);
+                return err;
+            }
+            if ((err = update_status(args, info, tid))) {
+                log_debug(args, "host update_status");
+                tid->abort(tid);
+                return err;
+            }
+            tid->commit(tid, 0);
             if (info->state_change) run_command(args,info);
             return 0;
         } 
@@ -264,9 +329,25 @@ int check_user(const abl_args *args, abl_info *info, time_t tm) {
 
         if (NULL != info->user) {
             log_debug(args, "Checking user %s", info->user);
-            check(args, info, tm);
-            update_status(args, info);
+            DB_TXN *tid;
+            int err = 0;
+            if ((err = info->dbenv->txn_begin(info->dbenv, NULL, &tid, 0)) != 0) {
+                log_db_error(err, "starting transaction");
+                return err;
+            }
+            err = check(args, info, tid, tm);
+            if (err) {
+                log_debug(args, "user check");
+                tid->abort(tid);
+                return err;
+            }
+            if ((err = update_status(args, info, tid))){
+                log_debug(args, "host update_status");
+                tid->abort(tid);
+                return err;
+            }
             if (info->state_change) run_command(args,info);
+            tid->commit(tid, 0);
             return 0;
         }
     } else {
