@@ -23,14 +23,16 @@
 
 static int record_host(const abl_args *args, abl_info *info, time_t tm) {
     if (NULL != args->host_db) {
-        const char *rhost;
+        const void *rhost;
         int err;
 
-        if (err = pam_get_item(args->pamh, PAM_RHOST, (const void **) &rhost), PAM_SUCCESS != err) {
+        if (err = pam_get_item(args->pamh, PAM_RHOST, &rhost), PAM_SUCCESS != err) {
             log_pam_error(args, err, "getting PAM_RHOST");
             return err;
         }
         if (NULL != rhost) {
+            if (!*(char *) rhost || !strcmp(rhost, "127.0.0.1"))
+                return 0;
             info->subject = HOST;
             info->host = rhost;
             return record(args, info, tm, args->host_purge);
@@ -44,9 +46,9 @@ static int record_host(const abl_args *args, abl_info *info, time_t tm) {
 }
 
 static int record_user(const abl_args *args, abl_info *info, time_t tm) {
-    const char *user;
+    const void *user;
     int err;
-    if (err = pam_get_item(args->pamh, PAM_USER, (const void **) &user), PAM_SUCCESS != err) {
+    if (err = pam_get_item(args->pamh, PAM_USER, &user), PAM_SUCCESS != err) {
         log_pam_error(args, err, "getting PAM_USER");
         return err;
     }
@@ -84,15 +86,18 @@ static int check_attempt(const abl_args *args, abl_info *info) {
         info->subject = HOST;
         err = check_host(args, info, tm);
         if (err) log_warning(args,"Failed to check host.");
-        if (info->state == BLOCKED ) {
-            /*No point in continuing since they've failed.*/
-            return 0;
-        }
+        short hostIsBlocked = 0;
+        if (info->state == BLOCKED )
+            hostIsBlocked = 1;
 
+        //we need to call check_user otherwise the block user script won't be called if needed.
         info->subject = USER;
         err = check_user(args, info, tm);
         if (err) log_warning(args,"Failed to check user.");
         
+        //only the info->state field is used in "pam_sm_authenticate", so that is the only field we need to restore.
+        if (hostIsBlocked)
+            info->state = BLOCKED;
     }
     return 0;
 }
@@ -118,6 +123,8 @@ static void cleanup(pam_handle_t *pamh, void *data, int err) {
         db->close(db,0);
         db = context->info->hsdb;
         db->close(db,0);
+        DB_ENV *dbenv = context->info->dbenv;
+        dbenv->close(dbenv, 0);
         free(context->info);
         free(context);
     }
@@ -128,10 +135,11 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, cons
     abl_args *args;
     abl_info *info;
     abl_context *context;
-    DB *utdb;
-    DB *usdb;
-    DB *htdb;
-    DB *hsdb;
+    DB *utdb = NULL;
+    DB *usdb = NULL;
+    DB *htdb = NULL;
+    DB *hsdb = NULL;
+    DB_ENV *dbenv;
     int err = PAM_SUCCESS;
 
     /*log_debug(args, "pam_sm_authenticate(), flags=%08x", flags);*/
@@ -155,17 +163,19 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, cons
 
         /* We now keep the database open from the beginning to avoid the cost
          * of opening them repeatedly. */
-
-        err = dbopen(args, args->user_db, DB_TIME, &utdb);
+        err = dbenvironment(args->db_home, &dbenv);
+        if (err) goto psa_fail;
+        info->dbenv = dbenv;
+        err = dbopen(args, args->user_db, DB_TIME, dbenv, &utdb);
         if (err) goto psa_fail;
         info->utdb = utdb;
-        err = dbopen(args, args->user_db, DB_STATE, &usdb);
+        err = dbopen(args, args->user_db, DB_STATE, dbenv, &usdb);
         if (err) goto psa_fail;
         info->usdb = usdb;
-        err = dbopen(args, args->host_db, DB_TIME, &htdb);
+        err = dbopen(args, args->host_db, DB_TIME, dbenv, &htdb);
         if (err) goto psa_fail;
         info->htdb = htdb;
-        err = dbopen(args, args->host_db, DB_STATE, &hsdb);
+        err = dbopen(args, args->host_db, DB_STATE, dbenv, &hsdb);
         if (err) goto psa_fail;
         info->hsdb = hsdb;
 
@@ -201,10 +211,16 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, cons
 
 psa_fail:
     config_free(args);
-    utdb->close(utdb,0);
-    usdb->close(usdb,0);
-    htdb->close(htdb,0);
-    hsdb->close(hsdb,0);
+    if (utdb)
+        utdb->close(utdb,0);
+    if (usdb)
+        usdb->close(usdb,0);
+    if (htdb)
+        htdb->close(htdb,0);
+    if (hsdb)
+        hsdb->close(hsdb,0);
+    if (dbenv)
+        dbenv->close(dbenv, 0);
     free(args);
     free(info);
     free(context);
