@@ -85,13 +85,14 @@ typedef struct {
     state_enum subject;
     state_enum state;
     int state_change;
-    const char *user;
-    const char *host;
-    const char *service;
+    const void *user;
+    const void *host;
+    const void *service;
     DB   *utdb;
     DB   *usdb;
     DB   *htdb;
     DB   *hsdb;
+    DB_ENV *dbenv;
 
 } abl_info;
 
@@ -108,6 +109,7 @@ typedef struct {
     int             expose_account;
 
     /* Our args */
+    const char      *db_home;
     const char      *host_db;
     const char      *host_rule;
     long            host_purge;
@@ -163,16 +165,17 @@ int rule_matchperiods(const abl_args *args, const time_t *history, int histsz,
 
 /* functions from common.c */
 void die(const char *msg, ...);
-int check(const abl_args *args, abl_info *info, time_t tm);
+int check(const abl_args *args, abl_info *info, DB_TXN *tid, time_t tm);
 int check_user(const abl_args *args, abl_info *info, time_t tm); 
 int check_host(const abl_args *args, abl_info *info, time_t tm);
 int record(const abl_args *args, abl_info *info, time_t tm, long maxage);
-int dbopen(const abl_args *args, const char *dbname, const char *dbfile, DB **db);
-int dbget(DB *db, DBT *key, DBT *dbtdata);
+int dbopen(const abl_args *args, const char *dbname, const char *dbfile, DB_ENV *dbenv, DB **db);
+int dbenvironment(const char *home, DB_ENV **dbenv);
+int dbget(DB *db, DBT *key, DB_TXN *tid, DBT *dbtdata);
 int prepare_command(const abl_args *args, const char *cmd, const abl_info *info,
         char **string);
 int run_command(const abl_args *args, const abl_info *info);
-int update_status(const abl_args *args, abl_info *info);
+int update_status(const abl_args *args, abl_info *info, DB_TXN *tid);
 void make_key(DBT *dbt, const char *key);
 
 #endif /* __PAM_ABL_H */
