@@ -20,6 +20,7 @@
 
 #include "../../config.h"
 #include "pam_abl.h"
+#include <inttypes.h>
 #include <getopt.h>
 
 #define MAXNAMES 200
@@ -157,12 +158,14 @@ static int doshow(const abl_args *args, abl_info *info) {
     DB *db=NULL;
     DBT key, data;
     DBC *cursor;
-    int err;
+    int err = 0;
     int bsz = 0;
     int cnt = 0;
     char *buf = NULL, *thing = NULL;
     const char *rule;
     time_t now = time(NULL);
+	u_int32_t fail_count;  // Matches db.h
+
 
     if (args == NULL || info== NULL) {
         return 0;
@@ -186,8 +189,14 @@ static int doshow(const abl_args *args, abl_info *info) {
     memset(&key,  0, sizeof(key));
     memset(&data, 0, sizeof(data));
 
-    if (err = db->cursor(db, NULL, &cursor, 0), 0 != err) {
-        log_sys_error(err, "creating cursor");
+    DB_TXN *tid = NULL;
+    if ((err = info->dbenv->txn_begin(info->dbenv, NULL, &tid, 0)) != 0) {
+        log_db_error(err, "starting transaction");
+        goto doshow_fail;
+    }
+    
+    if (err = db->cursor(db, tid, &cursor, 0), 0 != err) {
+        log_db_error(err, "creating cursor");
         goto doshow_fail;
     }
 
@@ -199,7 +208,7 @@ static int doshow(const abl_args *args, abl_info *info) {
         if (DB_NOTFOUND == err) {
             break;
         } else if (0 != err) {
-            log_sys_error(err, "iterating cursor");
+            log_db_error(err, "iterating cursor");
             goto doshow_fail;
         }
 
@@ -217,7 +226,8 @@ static int doshow(const abl_args *args, abl_info *info) {
 
         memcpy(buf, key.data, key.size);
         buf[key.size] = '\0';
-        printf(PAD "%s (%ld)\n", buf, data.size / sizeof(time_t));
+		fail_count = data.size / sizeof(time_t);
+        printf(PAD "%s (%"PRIu32")\n", buf, fail_count);
         cnt++;
 
         tm = (time_t *) data.data;
@@ -242,6 +252,16 @@ static int doshow(const abl_args *args, abl_info *info) {
 
     /* Cleanup */
 doshow_fail:
+#if DB_VERSION_MAJOR < 5
+    if (cursor != NULL) 
+        cursor->c_close(cursor);
+#else
+    if (cursor != NULL) 
+        cursor->close(cursor);
+#endif
+    if (tid) {
+        tid->commit(tid, 0);
+    }
     free(buf);
     return err;
 }
@@ -253,7 +273,7 @@ static int dopurge(const abl_args *args, abl_info *info) {
     DBC *cursor;
     int err;
     long maxage = 0;
-    const char **thing;
+    const void **thing;
     time_t now = time(NULL);
 
     if ( args == NULL || info == NULL ) {
@@ -278,8 +298,13 @@ static int dopurge(const abl_args *args, abl_info *info) {
     memset(&key,  0, sizeof(key));
     memset(&data, 0, sizeof(data));
 
-    if (err = tdb->cursor(tdb, NULL, &cursor, 0), 0 != err) {
-        log_sys_error(err, "creating cursor");
+    DB_TXN *tid = 0;
+    if ((err = info->dbenv->txn_begin(info->dbenv, NULL, &tid, 0)) != 0) {
+        log_db_error(err, "starting transaction");
+        return err;
+    }
+    if (err = tdb->cursor(tdb, tid, &cursor, 0), 0 != err) {
+        log_db_error(err, "creating cursor");
         goto dopurge_fail;
     }
 
@@ -288,19 +313,19 @@ static int dopurge(const abl_args *args, abl_info *info) {
         if (DB_NOTFOUND == err) {
             break;
         } else if (0 != err) {
-            log_sys_error(err, "iterating cursor");
+            log_db_error(err, "iterating cursor");
             goto dopurge_fail;
         }
 
         if (rule_purge(&data, maxage, now)) {
             if (data.size == 0) {
-                err = ( cursor->c_del(cursor, 0) || sdb->del(sdb, NULL, &key, 0) ); 
+                err = ( cursor->c_del(cursor, 0) || sdb->del(sdb, tid, &key, 0) ); 
                 if (err) {
                     goto dopurge_fail;
                 }
                 info->state = CLEAR;
                 *thing = (char*) key.data;
-                update_status(args, info);
+                update_status(args, info, tid);
                 if (info->state_change) run_command(args,info);
 
             } else {
@@ -315,12 +340,19 @@ static int dopurge(const abl_args *args, abl_info *info) {
 dopurge_fail:
 #if DB_VERSION_MAJOR < 5
     if (cursor != NULL) 
-        cursor->c_close;
+        cursor->c_close(cursor);
 #else
     if (cursor != NULL) 
         cursor->close(cursor);
 #endif
 
+    if (tid) {
+        if (err) {
+            tid->abort(tid);
+        } else {
+            tid->commit(tid, 0);
+        }
+    }
     return err;
 }
 
@@ -328,11 +360,11 @@ static int whitelist(const abl_args *args, abl_info *info, const char **permit,
     DB *db;
     int err = 0;
     DBT key, data;
-    DBC *cursor;
+    DBC *cursor = NULL;
     int wild = 0;
     int n, hit, del = 0;
     int *done;
-    const char **thing;
+    const void **thing;
 
     if (args == NULL || info == NULL) {
         return 0;
@@ -365,11 +397,16 @@ static int whitelist(const abl_args *args, abl_info *info, const char **permit,
     memset(&data, 0, sizeof(data));
     //data.flags = DB_DBT_REALLOC;
 
+    DB_TXN *tid = 0;
+    if ((err = info->dbenv->txn_begin(info->dbenv, NULL, &tid, 0)) != 0) {
+        log_db_error(err, "starting transaction");
+        goto whitelist_fail;
+    }
     if (wild) {
         /* wildcard delete */
 
-        if (err = db->cursor(db, NULL, &cursor, 0), 0 != err) {
-            log_sys_error(err, "creating cursor");
+        if (err = db->cursor(db, tid, &cursor, 0), 0 != err) {
+            log_db_error(err, "creating cursor");
             goto whitelist_fail;
         }
 
@@ -378,7 +415,7 @@ static int whitelist(const abl_args *args, abl_info *info, const char **permit,
             if (DB_NOTFOUND == err) {
                 break;
             } else if (0 != err) {
-                log_sys_error(err, "iterating cursor");
+                log_db_error(err, "iterating cursor");
                 goto whitelist_fail;
             }
 
@@ -395,7 +432,7 @@ static int whitelist(const abl_args *args, abl_info *info, const char **permit,
                 }
                 info->state = CLEAR;
                 *thing = permit[n];
-                update_status(args, info);
+                update_status(args, info, tid);
                 if (info->state_change) run_command(args,info);
                 del++;
             }
@@ -405,14 +442,14 @@ static int whitelist(const abl_args *args, abl_info *info, const char **permit,
         for (n = 0; n < count; n++) {
             if (!done[n]) {
                 make_key(&key, permit[n]);
-                err = db->del(db, NULL, &key, 0);
+                err = db->del(db, tid, &key, 0);
                 if (0 == err) {
                     int m;
                     done[n] = 1;
                     del++;
                     info->state = CLEAR;
                     *thing = permit[n];
-                    update_status(args, info);
+                    update_status(args, info, tid);
                     if (info->state_change) run_command(args,info);
                     /* Slightly pedantic - knock out any duplicates so we don't flag an error when
                      * we can't delete them later. Arguably if you supply duplicate arguments you
@@ -452,7 +489,22 @@ static int whitelist(const abl_args *args, abl_info *info, const char **permit,
     /* Cleanup */
 whitelist_fail:
     free(done);
-
+#if DB_VERSION_MAJOR < 5
+    if (cursor != NULL) {
+        cursor->c_close(cursor);
+    }
+#else
+    if (cursor != NULL) {
+        cursor->close(cursor);
+    }
+#endif
+    if (tid) {
+        if (err) {
+            tid->abort(tid);
+        } else {
+        	tid->commit(tid, 0);
+        }
+    }
     return err;
 }
 
@@ -530,6 +582,7 @@ int main(int argc, char **argv) {
     DB *usdb;
     DB *htdb;
     DB *hsdb;
+    DB_ENV *dbenv;
 
     while (1) {
 
@@ -626,16 +679,20 @@ int main(int argc, char **argv) {
 
     /* Most everything should be set, and it should be safe to open the
      * databases. */
-    err = dbopen(&args, args.user_db, DB_TIME, &utdb);
+    err = dbenvironment(args.db_home, &dbenv);
+    if (err) 
+        return err;
+    info.dbenv = dbenv;
+    err = dbopen(&args, args.user_db, DB_TIME, dbenv, &utdb);
     if (err) return err;
     info.utdb = utdb;
-    err = dbopen(&args, args.user_db, DB_STATE, &usdb);
+    err = dbopen(&args, args.user_db, DB_STATE, dbenv, &usdb);
     if (err) return err;
     info.usdb = usdb;
-    err = dbopen(&args, args.host_db, DB_TIME, &htdb);
+    err = dbopen(&args, args.host_db, DB_TIME, dbenv, &htdb);
     if (err) return err;
     info.htdb = htdb;
-    err = dbopen(&args, args.host_db, DB_STATE, &hsdb);
+    err = dbopen(&args, args.host_db, DB_STATE, dbenv, &hsdb);
     if (err) return err;
     info.hsdb = hsdb;
 
@@ -724,12 +781,12 @@ int main(int argc, char **argv) {
     }
 
     config_free(&args);
-
     err = utdb->close(utdb,0);
     err |= usdb->close(usdb,0);
     err |= htdb->close(htdb,0);
     err |= hsdb->close(hsdb,0);
-
+    if (dbenv)
+        dbenv->close(dbenv, 0);
     return 0;
 }
 
